import datetime
from dataclasses import _MISSING_TYPE

from dateutil.parser import parse as parse_date
from inverter import dc2colanderjson
from inverter.common import dataclass_check_type, dataclass_get_type

from ...interfaces import IDataProvider, ISchema
from ..app import App
from ..storage.memorystorage import MemoryStorage
from ..types import datestr

_MARKER: list = []


class DictProvider(IDataProvider):
    def __init__(self, schema, data, storage):
        self.schema = schema
        self.data = data
        self.storage = storage
        self.changed = False

    def __getitem__(self, key):
        t = dataclass_check_type(
            self.schema.__dataclass_fields__[key], datetime.datetime
        )
        if t:
            data = self.data[key]
            if isinstance(data, str):
                return parse_date(data)
            return data
        if key not in self.data.keys():
            field = self.schema.__dataclass_fields__[key]
            if not isinstance(field.default, _MISSING_TYPE):
                return field.default
            if not isinstance(field.default_factory, _MISSING_TYPE):
                return field.default_factory()
            return None
        return self.data[key]

    def __setitem__(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t["metadata"]["validators"]:
            v(schema=self.schema, request=self.storage.request, field=key, value=value)
        if t["type"] == datetime.datetime:
            if value and not isinstance(value, datetime.datetime):
                value = parse_date(value)
        if t["type"] == datetime.date:
            if value and not isinstance(value, datetime.date):
                value = parse_date(value).date()
        self.data[key] = value
        self.changed = True

    def __delitem__(self, key):
        del self.data[key]
        self.changed = True

    def setdefault(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t["metadata"]["validators"]:
            v(value)
        r = self.data.setdefault(key, value)
        self.changed = True
        return r

    def get(self, key, default=_MARKER):
        if default is _MARKER:
            return self.data.get(key)
        return self.data.get(key, default)

    def set(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t["metadata"]["validators"]:
            v(value)
        self.data[key] = value
        self.changed = True

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def as_dict(self):
        result = {}
        for k, v in self.data.items():
            result[k] = v
        return result

    def as_json(self):
        cschema = dc2colanderjson.convert(self.schema, request=self.storage.request)
        return cschema().serialize(self.as_dict())


@App.dataprovider(schema=ISchema, obj=dict, storage=MemoryStorage)
def get_dataprovider(schema, obj, storage):
    return DictProvider(schema, obj, storage)


@App.jsonprovider(obj=DictProvider)
def get_jsonprovider(obj):
    return obj.as_json()




import reg


class TypeRegistry(object):

    def __init__(self):
        self.types = []
        self.schema_name = {}

    def register_type(self, name, schema):
        if name not in self.types:
            self.types.append(name)
        self.schema_name[schema] = name

    def get_typeinfo(self, name, request):
        try:
            factory = request.app.get_typeinfo_factory(name)
        except NotImplementedError:
            factory = None

        if factory is None:
            raise KeyError('No type info registered for %s' % name)

        result = factory(request)  # self.typeinfo_factories[name](request)
        result['name'] = name
        return result

    def get_typeinfos(self, request):
        res = {}
        for k in self.types:
            res[k] = self.get_typeinfo(k, request)
        return res

    def get_typeinfo_by_schema(self, schema, request):
        name = self.schema_name.get(schema, None)
        if name is None:
            raise KeyError('No type info registered for %s' % schema)
        return self.get_typeinfo(name, request)



import argparse
import importlib
import os
import sys

import click
import morpfw
import reg
from cryptography.fernet import Fernet
from jinja2 import Template


@reg.dispatch(reg.match_class("app", lambda app_cls: app_cls))
def get_settings_opts(app_cls):
    raise NotImplementedError()


@get_settings_opts.register(app=morpfw.BaseApp)
def default_get_settings_opts(app_cls):
    proj_name = app_cls.__module__.split(".")[0]

    dburl = click.prompt(
        "Application DB URL",
        default="postgresql://postgres:postgres@localhost:5432/%s" % proj_name,
    )
    blobstorage_url = click.prompt(
        "Blobstorage URL", default=r"fsblob://%(here)s/blobstorage"
    )
    beaker_session_type = click.prompt("Beaker Session Type", default="ext:database")
    beaker_session_url = click.prompt(
        "Beaker Session URL",
        default="postgresql://postgres:postgres@localhost:5432/%s_cache" % proj_name,
    )
    beaker_cache_type = click.prompt("Beaker Cache Type", default=beaker_session_type)
    beaker_cache_url = click.prompt("Beaker Cache URL", default=beaker_session_url,)
    celery_broker_url = click.prompt(
        "Celery Broker URL", default="amqp://guest:guest@localhost:5672/%s" % proj_name
    )
    celery_result_backend = click.prompt(
        "Celery Result Backend",
        default="db+postgresql://postgres@localhost:5432/%s_cache" % proj_name,
    )
    return {
        "fernet_key": Fernet.generate_key().decode("utf-8"),
        "dburl": dburl,
        "blobstorage_url": blobstorage_url,
        "beaker_session_type": beaker_session_type,
        "beaker_session_url": beaker_session_url,
        "beaker_cache_type": beaker_cache_type,
        "beaker_cache_url": beaker_cache_url,
        "celery_broker_url": celery_broker_url,
        "celery_result_backend": celery_result_backend,
        "app_cls": "%s:%s" % (app_cls.__module__, app_cls.__name__),
        "app_module": app_cls.__module__.split(".")[0],
        "app_title": "Application",
    }


def genconfig(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--application",
        default=None,
        help="Application class, in format module:Class",
    )
    parser.add_argument("-o", "--output", default=None, help="Output file")
    args = parser.parse_args(argv)

    if args.application is None:
        if "MFW_APP" in os.environ:
            application = os.environ["MFW_APP"]
        else:
            parser.error("Application argument is required")
    else:
        application = args.application

    mod_name, cls_name = application.split(":")
    mod = importlib.import_module(mod_name)
    app_cls = getattr(mod, cls_name)

    for ac in app_cls.__mro__:
        c_mod = importlib.import_module(ac.__module__)
        mod_dir = os.path.dirname(c_mod.__file__)
        tmpl = os.path.join(mod_dir, "settings.yml.tmpl")

        if os.path.exists(tmpl):
            with open(tmpl, "r") as f:
                tmpl_str = f.read()

            template = Template(tmpl_str)
            out = template.render(get_settings_opts(app_cls))
            break

    if args.output is None:
        home_env = app_cls.home_env
        if home_env in os.environ:
            output = os.path.join(os.environ[home_env], "settings.yml")
        else:
            print("%s is not set. Using current directory" % home_env, file=sys.stderr)
            output = "settings.yml"
    else:
        output = args.output

    if os.path.exists(output):
        print("%s already exist" % output, file=sys.stderr)
        sys.exit(1)

    with open(output, "w") as of:
        of.write(out)

    print("Written %s" % output)



import copy
import errno
import importlib
import json
import os
import sys
from random import randint

import click
import yaml

from ..main import default_settings
from .generate_config import genconfig


def load_settings(settings_file, default=default_settings):
    if settings_file is None:
        if "MFW_APP" in os.environ:
            app_path = os.environ["MFW_APP"]
            modname, cname = app_path.split(":")
            mod = importlib.import_module(modname)
            app_cls = getattr(mod, cname)
            home_env = app_cls.home_env
            if home_env in os.environ:
                settings_file = os.path.join(os.environ[home_env], "settings.yml")

    dsettings = os.environ.get("MORP_SETTINGS", {})
    if settings_file is None:
        print(
            "WARNING: settings.yml not found, using default settings", file=sys.stderr
        )
        settings = dsettings
    elif not os.path.exists(settings_file):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), settings_file)
    else:
        raw_file = open(settings_file).read()
        raw_file = raw_file.replace(r"%(here)s", os.getcwd())
        settings = yaml.load(raw_file, Loader=yaml.Loader)

    s = copy.deepcopy(default)
    for k in settings.keys():
        if k in s.keys():
            for j, v in settings[k].items():
                s[k][j] = v
        else:
            s[k] = settings[k]

    settings = s
    os.environ["MORP_SETTINGS"] = json.dumps(settings)
    return settings


def load(settings_file: str = None, host: str = None, port: int = None):
    settings = load_settings(settings_file)

    if "application" not in settings:
        print("'application' section is required in settings")
        sys.exit(1)
    if "factory" not in settings["application"]:
        print("Missing application:factory entry in settings")
        sys.exit(1)
    if "class" not in settings["application"]:
        print("Missing application:class entry in settings")
        sys.exit(1)

    if "server" in settings:
        if not host:
            host = settings["server"].get("listen_host", "127.0.0.1")
        if not port:
            port = settings["server"].get("listen_port", 5432)

    sys.path.append(os.getcwd())

    factory_path = settings["application"]["factory"]
    mod, fname = factory_path.split(":")
    factory = getattr(importlib.import_module(mod), fname)

    app_path = settings["application"]["class"]
    mod, clsname = app_path.split(":")
    app_cls = getattr(importlib.import_module(mod), clsname)
    return {
        "factory": factory,
        "class": app_cls,
        "settings": settings,
        "host": host,
        "port": port,
    }


def confirmation_dialog(message="Are you sure you want to proceed?", token_length=6):

    token = ""
    for i in range(token_length):
        x = randint(0, 9)
        token += str(x)

    ans = input("%s \n(Please enter '%s' to confirm): " % (message, token))
    if ans.strip() != token:
        print("Invalid confirmation token")
        return False
    return True


@click.group()
@click.option("-s", "--settings", type=str, default=None, help="path to settings.yml")
@click.pass_context
def cli(ctx, settings):
    """Manage Morp application services"""
    ctx.ensure_object(dict)
    ctx.obj["settings"] = settings



from jsonschema import Draft4Validator


def validate(instance, schema):
    validator = Draft4Validator(schema)
    errors = [e for e in validator.iter_errors(instance)]
    field_errors = {}
    if errors:
        for e in errors:
            field = '.'.join(['form'] + list(e.path))
            field_errors.setdefault(field, [])
            field_errors[field].append(e.message)

    return field_errors if field_errors else None



import datetime
import hashlib
import json
import os
import typing
from pathlib import Path
from urllib.parse import urlparse
from uuid import uuid4

import morepath
import morpfw

from ..app import App
from .base import Blob, BlobStorage

# 1MB
WRITE_BUFF_SIZE = 1073741824


class FSBlob(Blob):
    def __init__(self, path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path

    def open(self):
        return open(self.path, "rb")

    def get_size(self):
        stat = os.stat(self.path)
        return stat.st_size


class FSBlobStorage(BlobStorage):
    def __init__(self, request: morepath.Request, path: str):
        self.path = path
        self.request = request

    def _uuid_path(self, uuid):
        first = uuid[:2]
        second = uuid[2:4]
        return os.path.join(self.path, first, second)

    def _meta_path(self, obj_path):
        return "%s.metadata.json" % obj_path

    def put(
        self,
        fileobj: typing.BinaryIO,
        filename: str,
        mimetype: typing.Optional[str] = None,
        size: typing.Optional[int] = None,
        encoding: typing.Optional[str] = None,
        uuid: typing.Optional[str] = None,
    ) -> FSBlob:

        if uuid is None:
            uuid = uuid4().hex

        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)
        Path(self._uuid_path(uuid)).mkdir(parents=True, exist_ok=True)

        fileobj.seek(0)
        hash = hashlib.sha256()
        with open(obj_path, "wb") as o:
            while True:
                data = fileobj.read(WRITE_BUFF_SIZE)
                if not data:
                    break
                hash.update(data)
                o.write(data)

        meta = {
            "uuid": uuid,
            "filename": filename,
            "mimetype": mimetype,
            "size": size,
            "encoding": encoding,
            "sha256sum": hash.hexdigest()
        }
        with open(meta_path, "w") as mo:
            mo.write(json.dumps(meta))

        meta["path"] = obj_path
        return FSBlob(**meta)

    def get(self, uuid: str) -> typing.Optional[Blob]:
        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)

        if not os.path.exists(meta_path):
            return None

        meta: dict = {}
        with open(meta_path) as mo:
            meta = json.loads(mo.read())

        meta["path"] = obj_path
        return FSBlob(**meta)

    def delete(self, uuid: str):
        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)

        if os.path.exists(obj_path):
            os.unlink(obj_path)

        if os.path.exists(meta_path):
            os.unlink(meta_path)


@App.blobstorage_factory("fsblob")
def get_fsblobstorage(request, uri):
    parsed = urlparse(uri)
    home_env = request.app.home_env
    if home_env not in os.environ:
        if 'MORP_WORKDIR' in os.environ:
            warnings.warn("MORP_WORKDIR environment is deprecated, use %s" %
                        home_env, DeprecationWarning)
            home_env = 'MORP_WORKDIR'

    if parsed.netloc == "":
        path = parsed.path
    else:
        path = os.path.join(
            os.environ.get(home_env, os.getcwd()),
            parsed.netloc,
            *parsed.path.split("/")
        )
    return FSBlobStorage(request, path)



import colander
from cryptography.fernet import Fernet


class EncryptedExportField(colander.String):
    """ 
    Serialize non-encrypted appstruct into encrypted cstruct. 
    """

    def __init__(self, fernet_key, *args, **kwargs):
        self.fernet_key = fernet_key
        self.fernet = Fernet(fernet_key)
        super().__init__(*args, **kwargs)

    def serialize(self, node, appstruct):
        v = super().serialize(node, appstruct)
        if v is colander.null:
            return v
        if v.strip():
            if v == "data":
                raise Exception()
            return self.fernet.encrypt(v.encode("utf8")).decode("utf8")
        return colander.null

    def deserialize(self, node, cstruct):
        v = super().deserialize(node, cstruct)
        if v is colander.null:
            return v
        # encrypt
        if v.strip():
            v = self.fernet.decrypt(v.encode("utf8")).decode("utf8")
            return v
        return colander.null


class EncryptedStoreField(colander.String):
    """ 
    Deserialize non-encrypted cstruct into encrypted appstruct. 
    """

    def __init__(self, fernet_key, *args, **kwargs):
        self.fernet_key = fernet_key
        self.fernet = Fernet(fernet_key)
        super().__init__(*args, **kwargs)

    def serialize(self, node, appstruct):
        """ Decrypt appstruct """
        v = super().serialize(node, appstruct)
        if v is colander.null:
            return v
        if v.strip():
            v = self.fernet.decrypt(v.encode("utf8")).decode("utf8")
            return v
        return colander.null

    def deserialize(self, node, cstruct):
        """ Encrypt cstruct """
        v = super().deserialize(node, cstruct)
        if v.strip():
            return self.fernet.encrypt(v.encode("utf8")).decode("utf8")
        return colander.null



import sqlalchemy as sa
import sqlalchemy.orm as saorm
import sqlalchemy_jsonfield as sajson
from morpfw.crud.storage.sqlstorage import GUID, Base
from sqlalchemy.orm import relationship


class User(Base):

    __tablename__ = "authmanager_users"

    username = sa.Column(sa.String(length=256))
    email = sa.Column(sa.String)
    password = sa.Column(sa.String(length=1024))
    source = sa.Column(sa.String(length=1024))
    nonce = sa.Column(sa.String(length=24))
    timezone = sa.Column(sa.String(length=124))
    is_administrator = sa.Column(sa.Boolean())


class APIKey(Base):

    __tablename__ = "authmanager_apikey"

    userid = sa.Column(GUID)
    name = sa.Column(sa.String(length=256))
    api_identity = sa.Column(sa.String(length=64))
    api_secret = sa.Column(sa.String(length=128))


class Group(Base):

    __tablename__ = "authmanager_groups"

    parent = sa.Column(sa.String(length=256))
    groupname = sa.Column(sa.String(length=256))
    memberships = relationship("Membership", cascade="all")


class Membership(Base):

    __tablename__ = "authmanager_membership"

    group_id = sa.Column(sa.ForeignKey("authmanager_groups.id"))
    user_id = sa.Column(sa.ForeignKey("authmanager_users.id"))
    sa.UniqueConstraint("group_id", "user_id", "deleted")

    roles_assignment = relationship("RoleAssignment", cascade="all")


class RoleAssignment(Base):

    __tablename__ = "authmanager_roleassignment"

    membership_id = sa.Column(sa.ForeignKey("authmanager_membership.id"))
    rolename = sa.Column(sa.String)
    sa.UniqueConstraint("membership_id", "role_id", "deleted")



import os

import morpfw
import requests
import webob
import yaml
from morpfw.app import SQLApp
from morpfw.authz.pas import DefaultAuthzPolicy
from morpfw.oauth import OAuthRoot
from morpfw.permission import All
from morpfw.sql import Base
from oauthlib.oauth2 import BackendApplicationClient, WebApplicationClient

from ..common import WebTestOAuth2Session, create_admin, get_client
from .test_auth import _test_authentication, login, logout


class Protected(All):
    pass


class App(SQLApp, DefaultAuthzPolicy):
    pass


App.hook_auth_models()


class AppRoot(object):
    def __init__(self, request) -> None:
        self.request = request


@App.path(model=AppRoot, path="/")
def get_approot(request):
    return AppRoot(request)


@App.permission_rule(model=AppRoot, permission=All)
def oauth_permission_rule(identity, context, permission):
    return True


@App.json(model=AppRoot, request_method="HEAD")
def head(context, request):
    return {"status": "success"}


@App.path(model=OAuthRoot, path="/oauth2")
def get_authroot(request):
    return OAuthRoot(request)


@App.json(model=AppRoot, name="protected-view", permission=Protected)
def get_protected_view(context, request):
    return {"status": "success"}


def test_oauth_sqlstorage(pgsql_db):
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
    config_file = os.path.join(os.path.dirname(__file__), "test_oauth-settings.yml")

    with open(config_file) as cf:
        settings = yaml.safe_load(cf.read())

    with morpfw.request_factory(settings) as request:
        Base.metadata.create_all(bind=request.db_session.bind)
        create_admin(request, "admin", "password", "admin@localhost.localdomain")

    with morpfw.request_factory(settings) as request:
        users = request.get_collection("morpfw.pas.user")
        user = users.get_by_username("admin")
        apikeys = request.get_collection("morpfw.pas.apikey")
        apikey = apikeys.create({"userid": user.uuid, "name": "test"})
        client_secret = apikey.generate_secret()
        client_id = apikey["api_identity"]

    c = get_client(config_file)

    client = BackendApplicationClient(client_id=client_id)
    oauth = WebTestOAuth2Session(c, client=client)

    token = oauth.fetch_token(
        token_url="/oauth2/token", client_id=client_id, client_secret=client_secret,
    )
    result = oauth.get("/+protected-view")

    # client = WebApplicationClient(client_id=client_id)
    # weboauth = WebTestOAuth2Session(c, client=client)

    # login(c, "admin", "password")
    # auth_url, state = weboauth.authorization_url("/oauth2/authorize")
    # r = c.get(auth_url)
    # r = c.post(auth_url, {"scopes": ["home"]})
    # logout(c)
    # token = weboauth.fetch_token(
    #    token_url="/oauth2/token",
    #    client_secret=client_secret,
    #    authorization_response=r.location,
    # )

    # result = weboauth.get("/+protected-view")

    # weboauth.refresh_token("/oauth2/token", refresh_token=token["refresh_token"])



from .app import App
from .model import Collection, Model


class MetalinkProvider(object):
    def __init__(self, request):
        self.request = request
        self.app = request.app

    def link(self, obj, **kwargs) -> dict:
        raise NotImplementedError

    def resolve(self, link) -> str:
        raise NotImplementedError


class CollectionMetalinkProvider(MetalinkProvider):
    def link(self, obj, view_name=None, **kwargs) -> dict:
        typeinfo = self.app.get_typeinfo_by_schema(obj.schema, self.request)
        return {
            "type": "morpfw.collection",
            "resource_type": typeinfo["name"],
            "view_name": view_name,
        }

    def resolve(self, link) -> str:
        col = self.request.get_collection(link["resource_type"])
        return self.request.link(col)


class ModelMetalinkProvider(MetalinkProvider):
    def link(self, obj, view_name=None, **kwargs) -> dict:
        typeinfo = self.app.get_typeinfo_by_schema(obj.schema, self.request)
        return {
            "type": "morpfw.model",
            "resource_type": typeinfo["name"],
            "uuid": obj.uuid,
            "view_name": view_name,
        }

    def resolve(self, link) -> str:
        col = self.request.get_collection(link["resource_type"])
        return self.request.link(col.get_by_uuid(link["uuid"]))


@App.metalink(name="morpfw.collection", model=Collection)
def get_collection_metalink(request):
    return CollectionMetalinkProvider(request)


@App.metalink(name="morpfw.model", model=Model)
def get_model_metalink(request):
    return ModelMetalinkProvider(request)




import abc
from ....interfaces import IModel, IStorageBase
from typing import Optional, Union, List, Sequence


class IUserModel(IModel):
    pass


class IGroupModel(IModel):
    pass


class IUserStorage(IStorageBase):
    @abc.abstractmethod
    def create(self, collection, data) -> IUserModel:
        """
        Create group
        """
        raise NotImplementedError

    @abc.abstractmethod
    def search(
        self,
        collection,
        query: Optional[dict] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> List[IUserModel]:
        """return search result based on specified rulez query"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_id(self, collection, id) -> IUserModel:
        """return model from internal ID"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, collection, uuid) -> IUserModel:
        """return model from GUID"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, collection, identifier, data: dict):
        """update model with values from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, collection, identifier, model: IUserModel):
        """delete model data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_userid(self, collection, model: IUserModel) -> str:
        """return the userid that will be used for internal references"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_email(self, collection, email: str) -> IUserModel:
        """return UserModel from email address"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_userid(self, collection, userid: str) -> IUserModel:
        """return UserModel from userid"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_username(self, collection, username: str) -> IUserModel:
        """return UserModel from username"""
        raise NotImplementedError

    @abc.abstractmethod
    def change_password(self, collection, userid: str, new_password: str):
        """update user's password"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_user_groups(self, collection, userid: str) -> IGroupModel:
        """get groups which the userid is a member of"""
        raise NotImplementedError

    @abc.abstractmethod
    def validate(self, collection, userid: str, password: str) -> bool:
        """validate userid's password"""
        raise NotImplementedError


class IGroupStorage(IStorageBase):
    @abc.abstractmethod
    def create(self, data):
        """
        Create group
        """
        raise NotImplementedError

    @abc.abstractmethod
    def search(
        self,
        query: Optional[dict] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> Sequence[IGroupModel]:
        """return search result based on specified rulez query"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_id(self, id) -> IGroupModel:
        """return model from internal ID"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, uuid) -> IGroupModel:
        """return model from GUID"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, identifier, data: dict):
        """update model with values from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, identifier, model: IGroupModel):
        """delete model data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_user_by_userid(self, userid: str, as_model: bool = True) -> IUserModel:
        """
        get user by its userid, and return
        as UserModel if as_model is True
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_user_by_username(self, username: str, as_model: bool = True) -> IUserModel:
        """
        get user by its username, and return
        as UserModel if as_model is True
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_members(self, groupname: str) -> List[IUserModel]:
        """get list of members of specified group"""
        raise NotImplementedError

    @abc.abstractmethod
    def add_group_members(self, groupname: str, userids: List[str]):
        """add userids into a group"""
        raise NotImplementedError

    @abc.abstractmethod
    def remove_group_members(self, groupname: str, userids: List[str]):
        """remove userids from group"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_group_user_roles(self, groupname: str, userid: str) -> List[str]:
        """get roles of userid in group"""
        raise NotImplementedError

    @abc.abstractmethod
    def grant_group_user_role(self, groupname: str, userid: str, rolename: str):
        """grant userid role in group"""
        raise NotImplementedError

    @abc.abstractmethod
    def revoke_group_user_role(self, groupname: str, userid: str, rolename: str):
        """revoke userid role in group"""
        raise NotImplementedError



import time

import morpfw
import pytest

# lets setup a skeleton app


class App(morpfw.BaseApp):
    pass


class Root(object):
    pass


@App.path(model=Root, path="")
def get_root(request):
    return Root()


# this view will dispatch the event


@App.json(model=Root)
def index(context, request):
    subs = request.async_dispatch("test_signal", obj={"data": 10})
    res = []
    for s in subs:
        res.append(s.get())
    return res


# these are the async handlers


@App.async_subscribe("test_signal")
def handler1(request_options, obj):
    # request_options contain parameters for instantiating a request
    with morpfw.request_factory(**request_options) as request:
        obj["handler"] = "handler1"
        obj["data"] += 1
        return obj


@App.async_subscribe("test_signal")
def handler2(request_options, obj):
    with morpfw.request_factory(**request_options) as request:
        obj["handler"] = "handler2"
        obj["data"] += 5
        return obj



import os
import sys

from setuptools import find_packages, setup

IS_RTD = os.environ.get("READTHEDOCS", None)

version = "0.4.0b12"

long_description = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()

install_requires = [
    "morepath==0.19",
    "alembic",
    "rulez>=0.1.4,<0.2.0",
    "inverter>=0.1.0<0.2.0",
    "more.cors",
    "celery",
    "redis",
    "jsl",
    "pyyaml>=4.2b1",
    "more.jsonschema",
    "sqlalchemy",
    "sqlalchemy_utils",
    "more.signals",
    "DateTime",
    "transitions",
    "jsonpath_ng",
    "python-dateutil",
    "more.jwtauth",
    "more.itsdangerous",
    "sqlsoup",
    "celery",
    "gunicorn",
    "itsdangerous",
    "pyyaml",
    "passlib",
    "jsonschema",
    "more.transaction",
    "zope.sqlalchemy",
    "python-dateutil",
    "more.cors",
    "sqlalchemy_jsonfield",
    "sqlsoup",
    "celery",
    "gunicorn",
    "itsdangerous",
    "pyyaml",
    "passlib",
    "jsonschema",
    "more.transaction",
    "zope.sqlalchemy",
    "more.basicauth",
    "cryptography",
    "elasticsearch>7.0.0,<8.0.0",
    "pamela",
    "click",
    "cookiecutter",
    "eventlet",
    "wsgigzip",
    "psycopg2",
    "colander",
    "deform",
    "more.chameleon",
    "more.static",
    "RestrictedPython",
    "beaker",
    "zstandard",
    "oauthlib[signedtoken]",
    "requests-oauthlib",
]

if IS_RTD is None:
    install_requires.append("python-ldap")

setup(
    name="morpfw",
    version=version,
    description="Web framework based on morepath",
    long_description=long_description,
    classifiers=[],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords="",
    author="Izhar Firdaus",
    author_email="izhar@kagesenshi.org",
    url="http://github.com/morpframework/morpfw",
    license="Apache-2.0",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        "test": [
            "nose",
            "webtest",
            "pytest",
            "pytest-html",
            "pytest_postgresql",
            "pytest_rabbitmq",
            "pytest-annotate",
            "pytest-cov",
            "pika",
            "mirakuru",
        ],
        "docs": ["sphinxcontrib-httpdomain", "sphinx-click"],
    },
    entry_points={
        "morepath": ["scan=morpfw"],
        "console_scripts": [
            "morpfw=morpfw.cli.main:main",
            "mfw-runmodule=morpfw.cli:run_module",
            "mfw-profilemodule=morpfw.cli:run_module_profile",
        ],
    },
)



import typing
import json

class UnprocessableError(Exception):
    def __init__(self, message=None):
        self.message = message or ""


class FormValidationError(Exception):
    def __init__(self, message):
        self.message = message


class FieldValidationError(Exception):
    def __init__(self, path, message):
        self.message = message
        self.path = path


class ValidationError(Exception):
    def __init__(
        self,
        field_errors: typing.Optional[typing.List[FieldValidationError]] = None,
        form_errors: typing.Optional[typing.List[FormValidationError]] = None,
    ):
        self.field_errors = field_errors or []
        self.form_errors = form_errors or []

        errobj = {
                'field_errors': [{'path': e.path, 'message': e.message} for e in self.field_errors],
                'form_errors': [e.message for e in self.form_errors]
        }
        message = "Schema validation error: {}".format(json.dumps(errobj, indent=4))
        super().__init__(message)
        


class AlreadyExistsError(Exception):
    def __init__(self, message):
        self.message = message


class StateUpdateProhibitedError(Exception):
    pass


class BlobStorageNotImplementedError(Exception):
    pass



import json
import os
import socket
import time
from multiprocessing import Process

import morepath
import morpfw
import morpfw.tests
import requests
import transaction
import webob
import yaml
from more.basicauth import BasicAuthIdentityPolicy
from more.jwtauth import JWTIdentityPolicy
from morpfw.authn.pas.exc import UserExistsError
from morpfw.cli import cli
from morpfw.main import create_admin as morpfw_create_admin
from morpfw.main import create_app
from morpfw.request import request_factory
from requests_oauthlib import OAuth2Session
from requests_oauthlib.oauth2_session import (
    InsecureTransportError,
    TokenExpiredError,
    TokenUpdated,
    is_secure_transport,
    log,
)
from webtest import TestApp as Client


def make_request(appobj):
    request = appobj.request_class(
        environ={
            "PATH_INFO": "/",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
            "SERVER_PROTOCOL": "HTTP/1.1",
        },
        app=appobj,
    )
    return request


def get_client(config="settings.yml", **kwargs):
    param = cli.load(config)
    morepath.scan(morpfw)
    request = request_factory(param["settings"], app_factory_opts=kwargs, scan=False)

    c = Client(request.environ["morpfw.wsgi.app"])
    c.mfw_request = request
    return c


def create_admin(client: Client, user: str, password: str, email: str):
    appobj = client.app
    morpfw_create_admin(client, user, password, email)
    transaction.commit()


def start_scheduler(app):
    settings = app._raw_settings
    hostname = socket.gethostname()
    ss = settings["configuration"]["morpfw.celery"]
    sched = app.celery.Beat(hostname="testscheduler.%s" % hostname, **ss)
    proc = Process(target=sched.run)
    proc.daemon = True
    proc.start()
    time.sleep(2)
    return proc


def start_worker(app):
    settings = app._raw_settings
    hostname = socket.gethostname()
    ss = settings["configuration"]["morpfw.celery"]
    worker = app.celery.Worker(hostname="testworker.%s" % hostname, **ss)
    proc = Process(target=worker.start)
    proc.daemon = True
    proc.start()
    time.sleep(2)
    return proc


class WebTestResponse(object):
    def __init__(self, response: webob.Response):
        self.wt_response = response

    @property
    def status_code(self):
        return self.wt_response.status_code

    @property
    def request(self):
        return self.wt_response.request

    @property
    def headers(self):
        return self.wt_response.headers

    @property
    def text(self):
        return self.wt_response.text


class FakeRequest(object):
    def __init__(self):
        self.headers = {}


class WebTestOAuth2Session(OAuth2Session):
    def __init__(self, wt_client, **kwargs):
        self.wt_client = wt_client
        super().__init__(**kwargs)

    def request(
        self,
        method,
        url,
        data=None,
        headers=None,
        auth=None,
        withhold_token=None,
        client_id=None,
        client_secret=None,
        **kwargs
    ):

        """Intercept all requests and add the OAuth 2 token if present."""
        if not is_secure_transport(url):
            raise InsecureTransportError()
        if self.token and not withhold_token:
            log.debug(
                "Invoking %d protected resource request hooks.",
                len(self.compliance_hook["protected_request"]),
            )
            for hook in self.compliance_hook["protected_request"]:
                log.debug("Invoking hook %s.", hook)
                url, headers, data = hook(url, headers, data)

            log.debug("Adding token %s to request.", self.token)
            try:
                url, headers, data = self._client.add_token(
                    url, http_method=method, body=data, headers=headers
                )
            # Attempt to retrieve and save new access token if expired
            except TokenExpiredError:
                if self.auto_refresh_url:
                    log.debug(
                        "Auto refresh is set, attempting to refresh at %s.",
                        self.auto_refresh_url,
                    )

                    # We mustn't pass auth twice.
                    auth = kwargs.pop("auth", None)
                    if client_id and client_secret and (auth is None):
                        log.debug(
                            'Encoding client_id "%s" with client_secret as Basic auth credentials.',
                            client_id,
                        )
                        auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
                    token = self.refresh_token(
                        self.auto_refresh_url, auth=auth, **kwargs
                    )
                    if self.token_updater:
                        log.debug(
                            "Updating token to %s using %s.", token, self.token_updater
                        )
                        self.token_updater(token)
                        url, headers, data = self._client.add_token(
                            url, http_method=method, body=data, headers=headers
                        )
                    else:
                        raise TokenUpdated(token)
                else:
                    raise

        log.debug("Requesting url %s using method %s.", url, method)
        log.debug("Supplying headers %s and data %s", headers, data)
        log.debug("Passing through key word arguments %s.", kwargs)
        if "json" in kwargs:
            f = getattr(self.wt_client, "%s_json" % method.lower())
            data = kwargs["json"]
        else:
            f = getattr(self.wt_client, method.lower())
        if auth:
            fr = FakeRequest()
            auth(fr)
            headers.update(fr.headers)
        resp = f(url, data, headers=headers)
        return WebTestResponse(resp)




import os

from morpfw.crud import pubsub

from ...app import SQLApp
from ...crud.storage.sqlstorage import SQLStorage
from ...sql import Base, construct_orm_model
from ..common import get_client, make_request
from .workflow_common import App as BaseApp
from .workflow_common import (DeliveryOrderCollection, DeliveryOrderModel,
                              DeliveryOrderSchema, KitchenOrderCollection,
                              KitchenOrderModel, KitchenOrderSchema,
                              OrderCollection, OrderModel, OrderSchema,
                              ProcessWorkflowCollection, ProcessWorkflowModel,
                              ProcessWorkflowSchema, run_test)


class App(BaseApp, SQLApp):
    pass


class OrderStorage(SQLStorage):
    model = OrderModel
    orm_model = construct_orm_model(
        schema=OrderModel.schema, metadata=Base.metadata, name="test_order"
    )


@App.storage(model=OrderModel)
def get_order_storage(model, request, blobstorage):
    return OrderStorage(request, blobstorage)


class KitchenOrderStorage(SQLStorage):
    model = KitchenOrderModel
    orm_model = construct_orm_model(
        schema=KitchenOrderModel.schema,
        metadata=Base.metadata,
        name="test_kitchen_order",
    )


@App.storage(model=KitchenOrderModel)
def get_kitchen_order_storage(model, request, blobstorage):
    return KitchenOrderStorage(request, blobstorage)


class DeliveryOrderStorage(SQLStorage):
    model = DeliveryOrderModel
    orm_model = construct_orm_model(
        schema=DeliveryOrderModel.schema,
        metadata=Base.metadata,
        name="test_delivery_order",
    )


@App.storage(model=DeliveryOrderModel)
def get_delivery_order_storage(model, request, blobstorage):
    return DeliveryOrderStorage(request, blobstorage)


class ProcessWorkflowStorage(SQLStorage):
    model = ProcessWorkflowModel
    orm_model = construct_orm_model(
        schema=ProcessWorkflowModel.schema,
        metadata=Base.metadata,
        name="test_process_workflow",
    )


@App.storage(model=ProcessWorkflowModel)
def get_process_wf_storage(model, request, blobstorage):
    return ProcessWorkflowStorage(request, blobstorage)


def test_workflow(pgsql_db):
    config = os.path.join(
        os.path.dirname(__file__), "test_workflow_sqlstorage-settings.yml"
    )
    c = get_client(config)
    request = make_request(c.app)
    Base.metadata.create_all(bind=request.db_session.bind)
    run_test(c)



from datetime import datetime

import colander
from inverter.common import dataclass_get_type

from ...interfaces import IDataProvider, ISchema
from ..app import App
from ..storage.elasticsearchstorage import ElasticSearchStorage
from .dictprovider import DictProvider


@App.dataprovider(schema=ISchema, obj=dict, storage=ElasticSearchStorage)
def get_dataprovider(schema, obj, storage):
    return ElasticSearchProvider(schema, obj, storage)


_MARKER = object()


def _deserialize(schema, key, value):
    res = schema[key].deserialize(value)
    if res in [colander.drop, colander.null]:
        return None
    return res


class ElasticSearchProvider(DictProvider):
    pass



from .model import APIKeyCollection, APIKeyModel, APIKeySchema
from ..app import App


def get_apikey_collection(request):
    storage = request.app.get_storage(APIKeyModel, request)
    return APIKeyCollection(request, storage)


def get_apikey(request, identifier):
    collection = get_apikey_collection(request)
    return collection.get(identifier)



import rulez
from morpfw.crud import errors as cruderrors
from morpfw.crud.storage.memorystorage import MemoryStorage

from .. import exc
from ..apikey.model import APIKeyModel, APIKeySchema
from ..app import App
from ..group.model import GroupCollection, GroupModel, GroupSchema
from ..user.model import UserCollection, UserModel, UserSchema
from .interfaces import IGroupStorage, IUserStorage

DB: dict = {"users": {}, "groups": {}, "rolemap": {}}


class UserMemoryStorage(MemoryStorage, IUserStorage):
    model = UserModel

    def get_userid(self, model):
        return model.uuid

    def get_by_email(self, collection, email):
        users = self.search(collection, rulez.field["email"] == email)
        if not users:
            return None
        return users[0]

    def get_by_userid(self, collection, userid):
        users = self.search(collection, rulez.field["uuid"] == userid)
        if not users:
            return None
        return users[0]

    def get_by_username(self, collection, username):
        users = self.search(collection, rulez.field["username"] == username)
        if not users:
            return None
        return users[0]

    def change_password(self, collection, userid, new_password):
        user = self.get_by_userid(collection, userid)
        user.data["password"] = new_password

    def get_user_groups(self, collection, userid):
        storage = self.request.app.get_storage(GroupModel, self.request)
        gcol = GroupCollection(self.request, storage)
        return storage.get_user_groups(userid, gcol)

    def validate(self, collection, userid, password):
        user = self.get_by_userid(collection, userid)
        return user.data["password"] == password


class APIKeyMemoryStorage(MemoryStorage):
    model = APIKeyModel


class GroupMemoryStorage(MemoryStorage, IGroupStorage):
    model = GroupModel

    def get_by_groupname(self, collection, groupname):
        res = self.search(collection, rulez.field["groupname"] == groupname)
        if res:
            return res[0]
        return None

    def get_user_by_userid(self, collection, userid, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        user_collection = UserCollection(self.request, user_storage)
        return user_storage.get_by_userid(user_collection, userid, as_model)

    def get_user_by_username(self, collection, username, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        user_collection = UserCollection(self.request, user_storage)
        return user_storage.get_by_username(user_collection, username)

    def get_user_groups(self, userid, collection):
        res = []
        for gid, group in self.datastore.items():
            if userid in group.data["xattrs"].get("members"):
                res.append(group)
        return res

    def get_members(self, collection, groupid):
        group = self.get(collection, groupid)
        userstorage = self.request.app.get_storage(UserModel, self.request)
        usercol = UserCollection(self.request, userstorage)
        res = []
        group.data.setdefault("xattrs", {})
        attrs = group.data["xattrs"]
        attrs.setdefault("members", [])
        for m in group.data["xattrs"].get("members"):
            res.append(userstorage.get_by_userid(usercol, m))
        return res

    def add_group_members(self, collection, groupid, userids):
        group = self.get(collection, groupid)
        group.data.setdefault("xattrs", {})
        attrs = group.data["xattrs"]
        attrs.setdefault("members", [])
        for u in userids:
            assert u is not None
            if u not in attrs["members"]:
                attrs["members"].append(u)
        group.data["xattrs"] = attrs

    def remove_group_members(self, collection, groupid, userids):
        group = self.get(collection, groupid)
        group.data.setdefault("xattrs", {})
        attrs = group.data["xattrs"]
        attrs.setdefault("members", [])
        for u in userids:
            if u in attrs["members"]:
                attrs["members"].remove(u)
        group.data["xattrs"] = attrs

    def get_group_user_roles(self, collection, groupid, userid):
        rolemap = DB["rolemap"]
        rolemap.setdefault(groupid, {})
        rolemap[groupid].setdefault(userid, [])
        return rolemap[groupid][userid]

    def grant_group_user_role(self, collection, groupid, userid, rolename):
        rolemap = DB["rolemap"]
        rolemap.setdefault(groupid, {})
        rolemap[groupid].setdefault(userid, [])
        if rolename not in rolemap[groupid][userid]:
            rolemap[groupid][userid].append(rolename)

    def revoke_group_user_role(self, collection, groupid, userid, rolename):
        rolemap = DB["rolemap"]
        rolemap.setdefault(groupid, {})
        rolemap[groupid].setdefault(userid, [])
        if rolename in rolemap[groupid][userid]:
            rolemap[groupid][userid].remove(rolename)


SINGLETON: dict = {}



import copy
import json
import os
import time

import morepath
import nose
import transaction
import yaml
from more.basicauth import BasicAuthIdentityPolicy
from more.jwtauth import JWTIdentityPolicy
from morpfw.authn.pas.app import App
from morpfw.authn.pas.user.model import GroupSchema, UserCollection, UserSchema
from morpfw.oauth import OAuthRoot
from oauthlib.oauth2 import BackendApplicationClient
from webtest import TestApp as Client

from ..common import WebTestOAuth2Session


def login(c, username, password="password"):

    r = c.post_json("/user/+login", {"username": username, "password": password})

    assert r.status_code == 200

    token = r.headers.get("Authorization").split()[1]

    c.authorization = ("Bearer", token)

    return r.json


def logout(c):
    c.authorization = None


def _test_authentication(c):
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

    r = c.get("/user/+login")

    # test schema access

    #    assert r.json['schema']['title'] == 'credential'

    ll = r.json["links"][0]
    assert ll["rel"] == "login"
    assert ll["type"] == "POST"

    r = c.post_json(ll["href"], {"username": "admin", "password": "password"})

    assert r.json == {"status": "success"}

    logout(c)

    # test wrong login
    r = c.post_json(
        "/user/+login",
        {"username": "invaliduser", "password": "invalidpassword"},
        expect_errors=True,
    )

    assert r.json == {
        "status": "error",
        "error": {"code": 401, "message": "Invalid Username / Password"},
    }

    r = c.post_json(
        "/user/+login",
        {"username": "admin", "password": "invalidpassword"},
        expect_errors=True,
    )

    assert r.json == {
        "status": "error",
        "error": {"code": 401, "message": "Invalid Username / Password"},
    }

    # from now on we login as admin

    login(c, "admin")

    # test refreshing token
    time.sleep(2)

    r = c.get("/self/+refresh_token")

    n = r.headers.get("Authorization").split()

    assert c.authorization[1] != n[1]

    c.authorization = ("Bearer", n[1])

    r = c.get("/user/admin")
    assert "password" not in r.json["data"]
    assert r.json["data"]["username"] == "admin"
    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__"
    ]
    assert r.json["data"]["state"] == "active"

    # query for nonexistent user

    r = c.get("/user/unknownuser", expect_errors=True)

    assert r.status_code == 404

    r = c.get("/user/")

    #    assert r.json['schema']['title'] == 'user'

    logout(c)

    # register new user, you shouldnt be allowed to if not logged in
    # as admin

    r = c.post_json(
        "/user/+register",
        {
            "username": "user1",
            "password": "password",
            "timezone": "UTC",
            "password_validate": "password",
        },
        expect_errors=True,
    )

    assert r.status_code == 403

    # register new user

    login(c, "admin")

    r = c.post_json(
        "/user/+register",
        {
            "username": "user1",
            "email": "user1@localhost.com",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
    )

    assert r.json == {"status": "success"}

    r = c.post_json(
        "/user/+register",
        {
            "username": "user2",
            "email": "user2@localhost.com",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
    )

    assert r.json == {"status": "success"}

    # fail if password is not string
    r = c.post_json(
        "/user/+register",
        {
            "username": "user3",
            "email": "user3@localhost.com",
            "timezone": "UTC",
            "password": {"hello": "world"},
        },
        expect_errors=True,
    )

    assert r.status_code == 422

    # fail if duplicate user

    r = c.post_json(
        "/user/+register",
        {
            "username": "user1",
            "email": "user1@localhost.com",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
        expect_errors=True,
    )

    assert r.status_code == 422
    assert r.json["status"] == "error"
    assert r.json["type"] == "UserExistsError"

    # fail if invalid email

    r = c.post_json(
        "/user/+register",
        {
            "username": "user4",
            "email": "user4",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
        expect_errors=True,
    )

    assert r.status_code == 422

    r = c.post_json(
        "/user/+register",
        {
            "username": "user4",
            "email": "user4@localhost",
            "timezone": "UTC",
            "password": "password",
            "password_validate": "password",
        },
        expect_errors=True,
    )

    assert r.status_code == 422

    r = c.get("/user/user1")

    assert r.json["data"]["username"] == "user1"
    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__"
    ]

    assert r.json["data"]["state"] == "active"
    assert "password" not in r.json["data"].keys()

    # attempt to login as user1
    login(c, "user1")

    # attempt to login as user1 with user1 email
    login(c, "user1@localhost.com")

    login(c, "admin")
    r = c.post_json("/user/user1/+statemachine", {"transition": "deactivate"})

    r = c.get("/user/user1")

    assert r.json["data"]["username"] == "user1"
    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__"
    ]
    assert r.json["data"]["state"] == "inactive"

    r = c.post_json(
        "/user/+login",
        {"username": "user1", "password": "password"},
        expect_errors=True,
    )

    assert r.status_code == 401

    r = c.post_json("/user/user1/+statemachine", {"transition": "activate"})

    r = c.get("/user/user1")

    assert r.json["data"]["username"] == "user1"
    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__"
    ]

    assert r.json["data"]["state"] == "active"

    login(c, "user1")

    # reject setting password through the update API

    r = c.patch_json("/user/user1/", {"password": "newpass"}, expect_errors=True)

    assert r.status_code == 422

    r = c.patch_json("/user/user1/", {"password": "newpass"}, expect_errors=True)

    assert r.status_code == 422

    r = c.patch_json("/user/user1/", {"username": "newusername"}, expect_errors=True)

    assert r.status_code == 422

    login(c, "admin")

    # admin should be allowed to set password without requiring current password
    r = c.post_json(
        "/user/user1/+change_password",
        {"new_password": "newpass", "new_password_validate": "newpass"},
    )

    assert r.status_code == 200

    login(c, "user1", "newpass")

    # changing password through user management API is limited only to admin
    r = c.post_json(
        "/user/user1/+change_password",
        {"new_password": "newpass", "new_password_validate": "newpass"},
        expect_errors=True,
    )

    assert r.status_code == 403

    # user require current password
    r = c.post_json(
        "/self/+change_password",
        {"new_password": "password", "new_password_validate": "password"},
        expect_errors=True,
    )

    assert r.status_code == 422

    # user require current password
    r = c.post_json(
        "/self/+change_password",
        {
            "password": "newpass",
            "new_password": "password",
            "new_password_validate": "password",
        },
        expect_errors=True,
    )

    assert r.status_code == 200

    login(c, "admin")

    r = c.get("/user/admin")

    admin_user = r.json

    # api keys
    r = c.post_json("/apikey/+generate", {"name": "samplekey", "password": "password"})

    key_identity = r.json["data"]["api_identity"]
    key_secret = r.json["data"]["api_secret"]
    key_uuid = r.json["data"]["uuid"]
    assert len(key_identity) == 43
    assert len(key_secret) == 43
    assert r.json["data"]["userid"] == admin_user["data"]["uuid"]

    r = c.get("/apikey/%s" % key_uuid)

    assert key_identity == r.json["data"]["api_identity"]
    assert "api_secret" not in r.json["data"]
    assert key_uuid == r.json["data"]["uuid"]

    # user1 shouldnt see admin's apikey

    login(c, "user1")

    r = c.get("/apikey/+search")

    assert len(r.json["results"]) == 0

    logout(c)

    # lets try deactivating user1 using API Oauth

    r = c.post_json(
        "/user/user1/+statemachine", {"transition": "deactivate"}, expect_errors=True
    )

    assert r.status_code == 403

    client = BackendApplicationClient(client_id=key_identity)
    oauth = WebTestOAuth2Session(c, client=client)
    oauth.fetch_token(
        token_url="/oauth2/token", client_id=key_identity, client_secret=key_secret,
    )

    r = oauth.post("/user/user1/+statemachine", json={"transition": "deactivate"})
    # r = c.post_json(
    #    "/user/user1/+statemachine",
    #    {"transition": "deactivate"},
    #    headers=[("X-API-KEY", ".".join([key_identity, key_secret]))],
    # )

    assert r.status_code == 200

    # activate back
    r = oauth.post("/user/user1/+statemachine", json={"transition": "activate"})
    # r = c.post_json(
    #    "/user/user1/+statemachine",
    #    {"transition": "activate"},
    #    headers=[("X-API-KEY", ".".join([key_identity, key_secret]))],
    # )

    assert r.status_code == 200

    login(c, "admin")

    r = c.get("/group/")

    #    assert r.json['schema']['title'] == 'group'

    r = c.post_json("/group/", {"groupname": "group1"})

    assert r.json["data"]["groupname"] == "group1"

    r = c.post_json("/group/", {"groupname": "group1"}, expect_errors=True)

    assert r.json["status"] == "error"
    assert r.json["type"] == "GroupExistsError"

    r = c.get("/group/group1")

    assert r.json["data"]["groupname"] == "group1"

    r = c.post_json(
        "/group/group1/+grant",
        {"mapping": [{"user": {"username": "user1"}, "roles": ["member"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.post_json(
        "/group/group1/+grant",
        {"mapping": [{"user": {"username": "dummyuser"}, "roles": ["member"]}]},
        expect_errors=True,
    )

    assert r.status_code == 422

    r = c.get("/user/user1")

    assert list(
        sorted([l["name"] for l in r.json["links"] if l["rel"] == "group"])
    ) == ["__default__", "group1"]

    r = c.get("/group/group1/+members")

    assert r.json["users"][0]["username"] == "user1"
    assert r.json["users"][0]["roles"] == ["member"]

    r = c.post_json(
        "/group/group1/+grant",
        {"mapping": [{"user": {"username": "user1"}, "roles": ["manager"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.get("/group/group1/+members")

    assert r.json["users"][0]["roles"] == ["member", "manager"]

    r = c.post_json(
        "/group/group1/+grant",
        {"mapping": [{"user": {"username": "user1"}, "roles": ["editor"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.get("/group/group1/+members")

    assert sorted(r.json["users"][0]["roles"]) == sorted(
        ["member", "editor", "manager"]
    )

    r = c.get("/user/user1/+roles")

    assert sorted(r.json["group1"]) == sorted(["member", "editor", "manager"])

    r = c.post_json(
        "/group/group1/+revoke",
        {"mapping": [{"user": {"username": "user1"}, "roles": ["manager"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.get("/group/group1/+members")

    assert r.json["users"][0]["roles"] == ["member", "editor"]

    r = c.post_json(
        "/group/group1/+revoke",
        {"mapping": [{"user": {"username": "user1"}, "roles": ["editor", "member"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.get("/group/group1/+members")

    assert len(r.json["users"]) == 0

    r = c.delete("/user/user1")

    assert r.json == {"status": "success"}

    r = c.get("/user/user1", expect_errors=True)

    assert r.status_code == 404

    logout(c)

    r = c.delete("/user/user1", expect_errors=True)

    login(c, "admin")

    r = c.get("/self")

    assert r.json["data"]["username"] == "admin"

    # Test delete group and ensure group is not return when user get info
    r = c.post_json("/group/", {"groupname": "group2"})

    assert r.json["data"]["groupname"] == "group2"

    # Grant role member to user2 in group2
    r = c.post_json(
        "/group/group2/+grant",
        {"mapping": [{"user": {"username": "user2"}, "roles": ["member"]}]},
    )

    assert r.json == {"status": "success"}

    r = c.get("/user/user2")

    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__",
        "group2",
    ]

    r = c.delete("/group/group2")

    assert r.json == {"status": "success"}

    r = c.get("/user/user2")

    assert list([l["name"] for l in r.json["links"] if l["rel"] == "group"]) == [
        "__default__"
    ]



from .model import UserCollection, UserModel
from .schema import UserSchema
from .path import get_user, get_user_collection
from ..app import App


@App.typeinfo(name='morpfw.pas.user',schema=UserSchema)
def get_typeinfo(request):
    return {
        'title': 'User',
        'description': 'User type',
        'schema': UserSchema,
        'collection': UserCollection,
        'collection_factory': get_user_collection,
        'model': UserModel,
        'model_factory': get_user,
        'internal': True
    }



import os

import yaml
from more.jwtauth import JWTIdentityPolicy
from morpfw.app import BaseApp
from morpfw.authn.pas.app import App
from morpfw.authn.pas.path import hook_auth_models
from morpfw.authn.pas.policy import MemoryStorageAuthApp as BaseAuthApp
from morpfw.authz.pas import DefaultAuthzPolicy

from ..common import create_admin, get_client
from .test_auth import _test_authentication


class MemoryStorageApp(BaseAuthApp, DefaultAuthzPolicy):
    pass


MemoryStorageApp.hook_auth_models()
MemoryStorageApp.hook_oauth_models()


def test_authentication_memorystorage():
    config = os.path.join(os.path.dirname(__file__), "settings-memorystorage.yml")
    c = get_client(config)
    create_admin(c.mfw_request, "admin", "password", "admin@localhost.localdomain")
    _test_authentication(c)



import typing
from dataclasses import dataclass, field

from morpfw.crud.schema import Schema

from ....validator.field import valid_identifier
from ..app import App


def valid_group_identifier(request, schema, field, value, mode=None):
    if value == "__default__":
        return
    return valid_identifier(request, schema, field, value, mode)


@dataclass
class GroupSchema(Schema):
    parent: typing.Optional[str] = field(
        default=None, metadata={"editable": False, "required": False}
    )
    groupname: typing.Optional[str] = field(
        default=None,
        metadata={
            "editable": False,
            "required": True,
            "validators": [valid_group_identifier],
        },
    )

    __unique_constraint__ = ["groupname", "deleted"]


@dataclass
class MemberSchema(Schema):
    users: typing.List[str] = field(default_factory=list, metadata={"required": True})



import os
from dataclasses import dataclass
from io import BytesIO

import sqlalchemy as sa
from morpfw import Collection, Model, Schema
from morpfw.crud.blobstorage.typeblobstorage import (
    TypeBlobStorage,
    TypeBlobStoreCollection,
    TypeBlobStoreMixin,
    TypeBlobStoreModel,
    TypeBlobStoreSchema,
)
from morpfw.crud.storage.sqlstorage import GUID, Base, SQLStorage

from ..common import get_client, make_request
from .crud_common import FSBLOB_DIR, run_jslcrud_test
from .test_sqlstorage import App as BaseApp
from .test_sqlstorage import BlobObjectModel


class App(BaseApp):
    pass


@dataclass
class BlobStoreSchema(TypeBlobStoreSchema):
    pass


class BlobStoreCollection(TypeBlobStoreCollection):
    schema = BlobStoreSchema


class BlobStoreModel(TypeBlobStoreModel):
    schema = BlobStoreSchema


class BlobStore(TypeBlobStoreMixin, Base):

    __tablename__ = "morpfw_test_blobobject"


class BlobStoreStorage(SQLStorage):
    model = BlobStoreModel
    orm_model = BlobStore


@App.path(model=BlobStoreCollection, path="blobs")
def blobstore_collection_factory(request):
    storage = BlobStoreStorage(request)
    return BlobStoreCollection(request, storage)


@App.path(model=BlobStoreModel, path="blobs/{identifier}")
def blobstore_model_factory(request, identifier):
    col = blobstore_collection_factory(request)
    return col.get(identifier)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return TypeBlobStorage(request, "blobstore")


@App.typeinfo(name="blobstore", schema=BlobStoreSchema)
def get_blobstore_typeinfo(request):
    return {
        "title": "BlobStore",
        "description": "BlobStore type",
        "schema": BlobStoreSchema,
        "collection": BlobStoreCollection,
        "collection_factory": blobstore_collection_factory,
        "model": BlobStoreModel,
        "model_factory": blobstore_model_factory,
        #
        # "collection_ui": BlobStoreCollectionUI,
        # "model_ui": BlobStoreModelUI,
        "internal": True
        #
    }


def test_sqlstorage_typeblobstorage(pgsql_db):
    config = os.path.join(
        os.path.dirname(__file__), "test_sqlstorage-typeblobstorage-settings.yml"
    )
    c = get_client(config)
    request = make_request(c.app)
    Base.metadata.create_all(bind=request.db_session.bind)
    run_jslcrud_test(c)



TASK_SUBMITTED: str = "morpfw.task_submitted"
TASK_STARTING: str = "morpfw.task_starting"
TASK_COMPLETED: str = "morpfw.task_completed"
TASK_FAILED: str = "morpfw.task_failed"
TASK_FINALIZED: str = "morpfw.task_finalized"
SCHEDULEDTASK_STARTING: str = "morpfw.scheduler.task_starting"
SCHEDULEDTASK_COMPLETED: str = "morpfw.scheduler.task_completed"
SCHEDULEDTASK_FAILED: str = "morpfw.scheduler.task_failed"
SCHEDULEDTASK_FINALIZED: str = "morpfw.scheduler.task_finalized"



from ..model import Collection
from ..app import App
from .base import SearchProvider


class StorageSearchProvider(SearchProvider):
    def search(self, query=None, offset=0, limit=None, order_by=None):
        return self.storage.search(self.context, query, offset, limit, order_by)


@App.searchprovider(model=Collection)
def get_searchprovider(context):
    return StorageSearchProvider(context)



import dectate
import reg
import celery
from uuid import uuid4


class CelerySubscriberAction(dectate.Action):

    app_class_arg = True

    config = {
        'celery_subscriber_registry': dict
    }

    def __init__(self, signal):
        self.signal = signal

    def identifier(self, celery_subscriber_registry, app_class):
        return str((app_class, uuid4().hex, self.signal))

    def perform(self, obj, celery_subscriber_registry, app_class):
        celery_subscriber_registry.setdefault(self.signal, [])
        celery_subscriber_registry[self.signal].append(obj)



import typing
from dataclasses import dataclass, field
from datetime import datetime
from pprint import pprint
from uuid import uuid4

import pytz

import colander
from inverter import dc2colander, dc2colanderjson

from ..interfaces import ISchema
from .app import App
from .errors import FieldValidationError, FormValidationError, ValidationError
from .relationship import BackReference, Reference


@dataclass
class BaseSchema(ISchema):
    @classmethod
    def validate(
        cls, request, data, deserialize=True, json=True, update_mode=False, **kwargs
    ):
        params = {}

        if not update_mode:
            if json:
                cschema = dc2colanderjson.convert(cls, request=request)
            else:
                cschema = dc2colander.convert(
                    cls, request=request, default_tzinfo=request.timezone()
                )

        else:
            if json:
                cschema = dc2colanderjson.convert(
                    cls, request=request, include_fields=data.keys(), mode="update",
                )
            else:
                cschema = dc2colander.convert(
                    cls,
                    request=request,
                    include_fields=data.keys(),
                    mode="update",
                    default_tzinfo=request.timezone(),
                )
        cs = cschema()
        # FIXME: need to pass context here
        cs = cs.bind(request=request, **kwargs)
        if not deserialize:
            # FIXME: can we skip this and immediately validate?

            data = cs.serialize(data)

        try:
            data = cs.deserialize(data)
        except colander.Invalid as e:
            errors = e.asdict()
            params["field_errors"] = [
                FieldValidationError(path=k, message=m) for k, m in errors.items()
            ]
        if params:
            raise ValidationError(**params)

        return data


@dataclass
class Schema(BaseSchema):

    id: typing.Optional[int] = field(
        default=None, metadata={"primary_key": True, "autoincrement": True}
    )
    uuid: typing.Optional[str] = field(
        default_factory=lambda: uuid4().hex,
        metadata={"format": "uuid", "index": True, "unique": True},
    )
    creator: typing.Optional[str] = field(
        default=None, metadata={"format": "uuid", "index": True, "length": 256}
    )
    created: typing.Optional[datetime] = field(
        default_factory=lambda: datetime.now(tz=pytz.UTC), metadata={"index": True}
    )
    modified: typing.Optional[datetime] = field(
        default_factory=lambda: datetime.now(tz=pytz.UTC), metadata={"index": True}
    )
    state: typing.Optional[str] = field(
        default=None, metadata={"index": True, "length": 64}
    )
    deleted: typing.Optional[datetime] = field(default=None, metadata={"index": True})

    blobs: typing.Optional[dict] = field(
        default_factory=dict, metadata={"exclude_if_empty": True}
    )
    xattrs: typing.Optional[dict] = field(
        default_factory=dict, metadata={"exclude_if_empty": True}
    )

    __unique_constraint__ = []  # type: ignore
    __references__ = []  # type: ignore
    __backreferences__ = []  # type: ignore
    __validators__ = []  # type: ignore
    __protected_fields__ = [
        "id",
        "blobs",
        "state",
        "xattrs",
        "modified",
        "created",
        "uuid",
        "creator",
        "deleted",
    ]


@App.identifierfield(schema=Schema)
def identifierfield(schema):
    return "uuid"


@App.default_identifier(schema=Schema)
def default_identifier(schema, obj, request):
    idfield = request.app.get_identifierfield(schema)
    if idfield == "uuid":
        obj[idfield] = uuid4().hex
    return obj[idfield]



import inspect
import threading
from datetime import datetime, timedelta

import pytz

from .interfaces import ICollection, IModel

threadlocal = threading.local()


class ModelMemoizer(object):
    def __init__(self, seconds: int = None):
        self.seconds = seconds

    def __call__(self, method):
        seconds = self.seconds

        def MemoizeWrapper(self, *args):
            klass = self.__class__
            nomemoize = self.request.environ.get("morpfw.nomemoize", False)
            if nomemoize:
                return method(self, *args)
            nomemoize = self.request.headers.get("X-MORP-NOMEMOIZE", None)
            if nomemoize is not None:
                return method(self, *args)
            if getattr(klass, "__memoize__", None) is None:

                memoizer = getattr(threadlocal, "mfw_classmemoize", None)
                if memoizer is None:
                    memoizer = {}
                    threadlocal.mfw_classmemoize = memoizer
                klass.__memoize__ = memoizer
            cachemgr = klass.__memoize__
            if isinstance(self, IModel):
                key = hash((self.__class__, method, self.uuid, args))
            elif isinstance(self, ICollection):
                key = hash((self.__class__, method, args))
            else:
                raise AssertionError(
                    "Memoization is only supported on IModel and ICollection instances"
                )
            cache = cachemgr.get(key, None)
            if cache:
                if isinstance(self, IModel) and cache["modified"] >= self["modified"]:
                    return cache["result"]

                if seconds:
                    if cache["modified"] >= (
                        datetime.now(tz=pytz.UTC) + timedelta(seconds=seconds)
                    ):
                        return cache["result"]

            result = method(self, *args)
            cachemgr[key] = {"result": result, "modified": datetime.now(tz=pytz.UTC)}
            return result

        MemoizeWrapper.__wrapped__ = method
        return MemoizeWrapper


class ModelRequestMemoizer(object):

    environ_key = "morpfw.memoize"

    def __init__(self, seconds: int = None):
        self.seconds = seconds

    def __call__(self, method):
        environ_key = self.environ_key
        seconds = self.seconds

        def RequestMemoizeWrapper(self, *args):
            nomemoize = self.request.environ.get("morpfw.nomemoize", False)
            if nomemoize:
                return method(self, *args)
            nomemoize = self.request.headers.get("X-MORP-NOMEMOIZE", None)
            if nomemoize is not None:
                return method(self, *args)
            self.request.environ.setdefault(environ_key, {})
            cachemgr = self.request.environ[environ_key]
            if isinstance(self, IModel):
                key = hash((self.__class__, method, self.uuid, args))
            elif isinstance(self, ICollection):
                key = hash((self.__class__, method, args))
            else:
                raise AssertionError(
                    "Memoization is only supported on IModel and ICollection instances"
                )
            cache = cachemgr.get(key, None)
            if cache:
                if isinstance(self, IModel) and cache["modified"] >= self["modified"]:
                    return cache["result"]

                if seconds:
                    if cache["modified"] >= (
                        datetime.now(tz=pytz.UTC) + timedelta(seconds=seconds)
                    ):
                        return cache["result"]

            result = method(self, *args)
            cachemgr[key] = {"result": result, "modified": datetime.now(tz=pytz.UTC)}
            return result

        RequestMemoizeWrapper.__wrapped__ = method
        return RequestMemoizeWrapper


memoize = ModelMemoizer
requestmemoize = ModelRequestMemoizer



import click
import morpfw

from .cli import cli, load


@cli.command(help="start web server")
@click.option("-h", "--host", default=None, help="Host")
@click.option("-p", "--port", default=None, type=int, help="Port")
@click.option("--prod", default=False, type=bool, is_flag=True, help="Production mode")
@click.option(
    "--workers", default=None, type=int, help="Number of workers to run in  prod mode"
)
@click.pass_context
def start(ctx, host, port, prod, workers):

    param = load(ctx.obj["settings"], host, port)
    if prod:
        morpfw.runprod(
            param["factory"](param["settings"]),
            settings=param["settings"],
            host=param["host"],
            port=param["port"],
            ignore_cli=True,
            workers=workers,
        )
    else:
        morpfw.run(
            param["factory"](param["settings"]),
            settings=param["settings"],
            host=param["host"],
            port=param["port"],
            ignore_cli=True,
        )



import typing
from dataclasses import dataclass

from ...app import App
from ..model import Model
from .base import XattrProvider

_marker = object()


class FieldXattrProvider(XattrProvider):

    field = "xattrs"

    def __getitem__(self, key):
        return self.context[self.field][key]

    def __setitem__(self, key, value):
        data = self.context[self.field]
        data[key] = value
        self.context.update({self.field: data})

    def __delitem__(self, key):
        data = self.context[self.field]
        del data[key]
        self.context.update({self.field: data})

    def get(self, key, default=_marker):
        if default is _marker:
            return self.context[self.field].get(key)
        return self.context[self.field].get(key, default)

    def as_json(self):
        return self.context.as_json().get(self.field, {})

    def as_dict(self):
        return self.context[self.field] or {}

    def update(self, newdata: dict):
        data = self.context[self.field] or {}
        data.update(newdata)
        self.context.update({self.field: data})



from ..permission import All as MFWAll

class All(MFWAll):
    pass

class Create(All):
    pass

class View(Create):
    pass

class Edit(All):
    pass

class Delete(All):
    pass

class Search(All):
    pass

class Aggregate(All):
    pass

class StateUpdate(MFWAll):
    pass



import os

import jsl
import morpfw.crud.signals as signals
from inverter import dc2pgsqla
from more.basicauth import BasicAuthIdentityPolicy
from more.transaction import TransactionApp
from morpfw.app import SQLApp
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from morpfw.crud.storage.pgsqlstorage import PgSQLStorage
from sqlalchemy import MetaData

from ..common import get_client, make_request
from .crud_common import FSBLOB_DIR
from .crud_common import App as BaseApp
from .crud_common import (
    BlobObjectCollection,
    BlobObjectModel,
    NamedObjectCollection,
    NamedObjectModel,
    ObjectCollection,
    ObjectModel,
    PageCollection,
    PageModel,
    PageSchema,
    run_jslcrud_test,
)

db_meta = MetaData()


class App(BaseApp, SQLApp):
    pass


class PageStorage(PgSQLStorage):
    model = PageModel


@App.path(model=PageCollection, path="pages")
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="pages/{identifier}")
def model_factory(request, identifier):
    col = collection_factory(request)
    return col.get(identifier)


@App.typeinfo(name="tests.page", schema=PageSchema)
def get_page_typeinfo(request):
    return {
        "title": "Page",
        "description": "Page type",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": collection_factory,
        "model": PageModel,
        "model_factory": model_factory,
    }


class ObjectStorage(PgSQLStorage):
    model = ObjectModel


@App.path(model=ObjectCollection, path="objects")
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path="objects/{identifier}")
def object_model_factory(request, identifier):
    col = object_collection_factory(request)
    return col.get(identifier)


class NamedObjectStorage(PgSQLStorage):
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path="named_objects")
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path="named_objects/{identifier}")
def namedobject_model_factory(request, identifier):
    col = namedobject_collection_factory(request)
    return col.get(identifier)


class BlobObjectStorage(PgSQLStorage):
    model = BlobObjectModel


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path="blob_objects")
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path="blob_objects/{identifier}")
def blobobject_model_factory(request, identifier):
    col = blobobject_collection_factory(request)
    return col.get(identifier)


def test_pgsqlstorage(pgsql_db):
    config = os.path.join(os.path.dirname(__file__), "test_pgsqlstorage-settings.yml")
    c = get_client(config)
    request = make_request(c.app)

    # create tables

    db_meta.bind = request.db_session.get_bind()
    for model in [PageModel, ObjectModel, NamedObjectModel, BlobObjectModel]:
        dc2pgsqla.convert(model.schema, db_meta)
    db_meta.create_all()

    run_jslcrud_test(c)



import os
import pprint

import jsl
import morpfw.crud.signals as signals
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from morpfw.app import SQLApp
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from morpfw.crud.model import Collection, Model
from morpfw.crud.storage.sqlstorage import GUID, Base, SQLStorage
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session

from ..common import get_client, make_request
from .crud_common import FSBLOB_DIR
from .crud_common import App as BaseApp
from .crud_common import (
    BlobObjectCollection,
    BlobObjectModel,
    NamedObjectCollection,
    NamedObjectModel,
    ObjectCollection,
    ObjectModel,
    PageCollection,
    PageModel,
    PageSchema,
    run_jslcrud_test,
)


class App(BaseApp, SQLApp):
    pass


class Page(Base):

    __tablename__ = "jslcrud_test_page"

    title = sa.Column(sa.String(length=1024), default="")
    body = sa.Column(sa.Text(), default="")
    publish_date = sa.Column(sa.Date())
    publish_datetime = sa.Column(sa.DateTime(timezone=True))
    value = sa.Column(sa.Integer)
    footer = sa.Column(sa.String(length=1024), default="")
    computed_title = sa.Column(sa.String(length=1024), default="")

class PageStorage(SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path="pages")
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="pages/{identifier}")
def model_factory(request, identifier):
    col = collection_factory(request)
    return col.get(identifier)


@App.typeinfo(name="tests.page", schema=PageSchema)
def get_page_typeinfo(request):
    return {
        "title": "Page",
        "description": "Page type",
        "schema": PageSchema,
        "collection": PageCollection,
        "collection_factory": collection_factory,
        "model": PageModel,
        "model_factory": model_factory,
    }


class Object(Base):

    __tablename__ = "jslcrud_test_object"

    body = sa.Column(sa.String(length=1024), default="")
    created_flag = sa.Column(sa.Boolean, default=False)
    updated_flag = sa.Column(sa.Boolean, default=False)
    attrs = sa.Column(sajson.JSONField)


class ObjectStorage(SQLStorage):
    model = ObjectModel
    orm_model = Object


@App.path(model=ObjectCollection, path="objects")
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path="objects/{identifier}")
def object_model_factory(request, identifier):
    col = object_collection_factory(request)
    return col.get(identifier)


class NamedObject(Base):

    __tablename__ = "jslcrud_test_namedobject"

    name = sa.Column(sa.String(length=1024), default="")
    body = sa.Column(sa.String(length=1024), default="")
    created_flag = sa.Column(sa.Boolean, default=False)
    updated_flag = sa.Column(sa.Boolean, default=False)


class NamedObjectStorage(SQLStorage):
    model = NamedObjectModel
    orm_model = NamedObject


@App.path(model=NamedObjectCollection, path="named_objects")
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path="named_objects/{identifier}")
def namedobject_model_factory(request, identifier):
    col = namedobject_collection_factory(request)
    return col.get(identifier)


class BlobObject(Base):

    __tablename__ = "jslcrud_test_blobobject"

    blobs = sa.Column(sajson.JSONField())


class BlobObjectStorage(SQLStorage):
    model = BlobObjectModel
    orm_model = BlobObject


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path="blob_objects")
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path="blob_objects/{identifier}")
def blobobject_model_factory(request, identifier):
    col = blobobject_collection_factory(request)
    return col.get(identifier)


def test_sqlstorage(pgsql_db):
    config = os.path.join(os.path.dirname(__file__), "test_sqlstorage-settings.yml")
    c = get_client(config)
    request = make_request(c.app)
    Base.metadata.create_all(bind=request.db_session.bind)
    run_jslcrud_test(c)
